import time
import traceback
import httpx
from .api_info import ApiInfo
from .base_client import BaseClient
from .helpers import *

class RecNetLogin(BaseClient):
    def __init__(
        self, 
        **kwargs
    ) -> None:
        BaseClient.__init__(self, **kwargs)
        # HTTP session
        self.__session = self.session if isinstance(self.session, httpx.Client) else httpx.Client()
       
    def __login(self) -> None:
        resp = self.__session.get(ApiInfo.URL)
        html = resp.content
        
        rvt = parse_rvt(html=html)
        body = create_body(username=self.username, password=self.password, rvt=rvt)
            
        resp = self.__session.post(ApiInfo.URL + ApiInfo.PARAMS, headers=ApiInfo.HEADERS, data=body, follow_redirects=True)
        try:
            self.bearer_token = parse_token(resp.url)
            self.decoded = decode_token(self.bearer_token)
        except TwoFactorAuthenticatorEnabled:
            if not self.prompt_2fa: raise TwoFactorAuthenticatorProcessDisabled
            self.__login_2fa()
        except Exception:  # Uncaught exception
            traceback.print_exc()
            return
        
    def __login_2fa(self) -> None:
        twofa_code = prompt_2fa()

        resp = self.__session.get(ApiInfo.TWOFA_URL)
        html = resp.content
        rvt = parse_rvt(html=html)
                
        body = create_twofa_body(twofa_code=twofa_code, rvt=rvt)

        resp = self.__session.post(ApiInfo.TWOFA_URL + ApiInfo.TWOFA_PARAMS, headers=ApiInfo.HEADERS, data=body, follow_redirects=True)

        try:
            self.bearer_token = parse_token(resp.url)
            self.decoded = decode_token(self.bearer_token)   
        except TwoFactorAuthenticatorEnabled:
            print(debug_text("Authenticator code is wrong!"))
            return self.__login_2fa() 

    def close(self):
        """
        Closes the HTTPX client
        """
        
        self.__session.close()

    def __renew_token(func):
        def wrapper(self, *args, **kwargs) -> func:        
            if not self.bearer_token or not self.decoded or int(time.time()) >= self.decoded['exp']:
                self.__login()
            return func(self, *args, **kwargs)
        return wrapper

    @__renew_token
    def get_token(self, include_bearer: bool = False) -> str:
        token = self.bearer_token
        if include_bearer: token = "Bearer " + token
        return token

    @__renew_token
    def get_decoded_token(self) -> dict:
        return self.decoded


import jwt
from bs4 import BeautifulSoup
from urllib.parse import urlparse, parse_qs, quote
from .api_info import ApiInfo
from .exceptions import *

def parse_token(url: str) -> tuple:
    redirect_url = str(url).replace("#", "?")  # Patch the URL
    if 'Lockout' in redirect_url: raise Lockout
    if 'LoginWith2fa' in redirect_url: raise TwoFactorAuthenticatorEnabled  # 2FA enabled! 
    if 'access_token' not in redirect_url: raise InvalidAccountCredentials  # Login unsuccessful!
    parsed_url = urlparse(redirect_url)
    bearer_token = parse_qs(parsed_url.query)['access_token'][0]  # Acquire the bearer token from the URL
    return bearer_token

def decode_token(token: str) -> dict:
    decoded = jwt.decode(token, options={"verify_signature": False})  # Decode the bearer token
    return decoded
            
def parse_rvt(html: str) -> str:
    soup = BeautifulSoup(html, features="html.parser")
    rvt = soup.find('input', {'name': '__RequestVerificationToken'}).get('value')  # Acquire the Request Verification Token
    return quote(rvt)

def create_body(username: str, password: str, rvt: str) -> str:
    body = ApiInfo.BODY.format(name=quote(username), password=quote(password), rvt=rvt)
    return body

def create_twofa_body(twofa_code: str, rvt: str) -> str:
    body = ApiInfo.TWOFA_BODY.format(twofa_code=twofa_code, rvt=rvt)
    return body

def prompt_2fa() -> str:
    twofa_code = ""
    while True:
        twofa_code = input(debug_text("Authenticator code: "))
        if not twofa_code: raise TwoFactorAuthenticatorCancelled
        if len(twofa_code) >= 6 and len(twofa_code) <= 7:
            return twofa_code
        print(debug_text("The Authenticator code must be at least 6 and at max 7 characters long."))

def debug_text(text: str) -> str:
    signature = "RecNetLogin – "
    return signature + text


"""
Exceptions for the RecNetLogin clients
"""

EXCEPTION_SIGNATURE = "Unable to login to RecNet – "

class Lockout(Exception):
    """Raised when the Rec Room account is locked out."""
    def __init__(self) -> None:
        super().__init__(EXCEPTION_SIGNATURE + "Rec Room account locked!")

class InvalidAccountCredentials(Exception):
    """Raised when the Rec Room account credentials are invalid."""
    def __init__(self) -> None:
        super().__init__(EXCEPTION_SIGNATURE + "Invalid Rec Room account credentials!")

class TwoFactorAuthenticatorCodeWrong(Exception):
    """Raised when the Authenticator code is wrong."""
    def __init__(self) -> None:
        super().__init__(EXCEPTION_SIGNATURE + "Wrong Authenticator code!")
        
class TwoFactorAuthenticatorEnabled(Exception):
    """Raised when the Rec Room account has 2FA enabled."""
    def __init__(self) -> None:
        super().__init__(EXCEPTION_SIGNATURE + "2FA enabled!")

class TwoFactorAuthenticatorCancelled(Exception):
    """Raised when the authentication process is cancelled."""
    def __init__(self) -> None:
        super().__init__(EXCEPTION_SIGNATURE + "2FA authentication cancelled!")

class TwoFactorAuthenticatorProcessDisabled(Exception):
    """Raised when the authentication process is disabled."""
    def __init__(self) -> None:
        super().__init__(EXCEPTION_SIGNATURE + "2FA authentication process disabled!")


class ApiInfo():
    URL: str = "https://auth.rec.net/Account/Login"
    TWOFA_URL: str = "https://auth.rec.net/Account/LoginWith2fa"
    BODY: str = "Input.Username={name}&Input.Password={password}&Input.RememberMe=true&button=login&__RequestVerificationToken={rvt}"
    TWOFA_BODY: str = "RememberMe=True&Input.TwoFactorCode={twofa_code}&__RequestVerificationToken={rvt}&Input.RememberMachine=true"
    PARAMS: str = "?ReturnUrl=%2Fconnect%2Fauthorize%2Fcallback%3Fclient_id%3Drecnet%26redirect_uri%3Dhttps%253A%252F%252Frec.net%252Fauthenticate%252Fdefault%26response_type%3Did_token%2520token%26scope%3Dopenid%2520rn.api%2520rn.notify%2520rn.match.read%2520rn.chat%2520rn.accounts%2520rn.auth%2520rn.link%2520rn.clubs%2520rn.rooms%26state%3Da01a9f406219471ebfa9602eb3abfa97%26nonce%3D99a86479c2f54ec5af90152184d2049c"
    TWOFA_PARAMS: str = "?ReturnUrl=%2Fconnect%2Fauthorize%2Fcallback%3Fclient_id%3Drecnet%26redirect_uri%3Dhttps%253A%252F%252Frec.net%252Fauthenticate%252Fdefault%26response_type%3Did_token%2520token%26scope%3Dopenid%2520rn.api%2520rn.notify%2520rn.match.read%2520rn.chat%2520rn.accounts%2520rn.auth%2520rn.link%2520rn.clubs%2520rn.rooms%26state%3D0b156b90d78249cd95cc54ed965cf705%26nonce%3D3a64f6e9e3324f3ca10dd0ad57fd83f0&RememberMe=True"
    HEADERS: dict = {"Content-Type": "application/x-www-form-urlencoded", "Connection": "keep-alive"}


import aiohttp
import time
import traceback
from .api_info import ApiInfo
from .base_client import BaseClient
from .helpers import *

class RecNetLoginAsync(BaseClient):
    def __init__(
        self, 
        **kwargs
    ) -> None:
        BaseClient.__init__(self, **kwargs)
        # HTTP session
        self.__session = self.session if isinstance(self.session, aiohttp.ClientSession) else aiohttp.ClientSession(headers=ApiInfo.HEADERS)
       
    async def __login(self) -> None:
        async with self.__session.get(ApiInfo.URL) as resp:
            html = await resp.text()
            rvt = parse_rvt(html=html)
            
        # Body with the account details
        body = create_body(username=self.username, password=self.password, rvt=rvt)
        
        async with self.__session.post(ApiInfo.URL + ApiInfo.PARAMS, data=body, allow_redirects=True, headers=ApiInfo.HEADERS) as resp:
            try:
                self.bearer_token = parse_token(resp.real_url)
                self.decoded = decode_token(self.bearer_token)
            except TwoFactorAuthenticatorEnabled:
                if not self.prompt_2fa: raise TwoFactorAuthenticatorProcessDisabled
                await self.__login_2fa()

    async def __login_2fa(self) -> None:
        twofa_code = prompt_2fa()

        async with self.__session.get(ApiInfo.TWOFA_URL) as resp:
            html = await resp.text()
            rvt = parse_rvt(html=html)
            
        body = create_twofa_body(twofa_code=twofa_code, rvt=rvt)

        async with self.__session.post(ApiInfo.TWOFA_URL + ApiInfo.TWOFA_PARAMS, headers=ApiInfo.HEADERS, data=body, allow_redirects=True) as resp:
            try:
                self.bearer_token = parse_token(resp.real_url)
                self.decoded = decode_token(self.bearer_token)   
            except TwoFactorAuthenticatorEnabled:
                print(debug_text("Authenticator code is wrong!"))
                return await self.__login_2fa() 
            except Exception:  # Uncaught exception
                traceback.print_exc()
                return

    async def close(self) -> None:
        """
        Closes the AIOHTTP client
        """
        
        await self.__session.close()

    def __renew_token(func):
        async def wrapper(self, *args, **kwargs) -> func:        
            if not self.bearer_token or not self.decoded or int(time.time()) >= self.decoded['exp']:
                await self.__login()
            return await func(self, *args, **kwargs)
        return wrapper

    @__renew_token
    async def get_token(self, include_bearer: bool = False) -> str:
        token = self.bearer_token
        if include_bearer: token = "Bearer " + token
        return token

    @__renew_token
    async def get_decoded_token(self) -> dict:
        return self.decoded


import aiohttp
import httpx

class BaseClient():
    def __init__(
        self, 
        username: str, 
        password: str,
        prompt_2fa: bool = False,
        session: aiohttp.ClientSession or httpx.Client = None,
        **kwargs
    ) -> None:
        # RecNet credentials
        self.username, self.password = username, password
        # HTTP session
        self.session = session
        # The bearer token
        self.bearer_token = ""
        # Decoded bearer token
        self.decoded = {}
        # Whether or not to prompt for 2fa code
        self.prompt_2fa = prompt_2fa